"""This package contains LayerGenerator definitions."""
from __future__ import annotations

from bqskit.passes.search.generators.simple import SimpleLayerGenerator

__all__ = ['SimpleLayerGenerator']
